#pragma clang diagnostic ignored "-Wmissing-prototypes"
#pragma clang diagnostic ignored "-Wmissing-braces"

#include <metal_stdlib>
#include <simd/simd.h>

using namespace metal;

template<typename T, size_t Num>
struct spvUnsafeArray
{
    T elements[Num ? Num : 1];
    
    thread T& operator [] (size_t pos) thread
    {
        return elements[pos];
    }
    constexpr const thread T& operator [] (size_t pos) const thread
    {
        return elements[pos];
    }
    
    device T& operator [] (size_t pos) device
    {
        return elements[pos];
    }
    constexpr const device T& operator [] (size_t pos) const device
    {
        return elements[pos];
    }
    
    constexpr const constant T& operator [] (size_t pos) const constant
    {
        return elements[pos];
    }
    
    threadgroup T& operator [] (size_t pos) threadgroup
    {
        return elements[pos];
    }
    constexpr const threadgroup T& operator [] (size_t pos) const threadgroup
    {
        return elements[pos];
    }
};

struct buffer_t
{
    spvUnsafeArray<float4x4, 50> u_Palatte;
    float4x4 u_Model;
    float4x4 u_TransposeInvModel;
    float4x4 u_MVP;
    float2 _Offset;
    float2 _Tiling;
    float _Rotation;
};

struct main0_out
{
    float2 v_uv0;
    float3 v_nDirWS;
    float3 v_posWS;
    float4 gl_Position [[position]];
};

struct main0_in
{
    float3 attPosition [[attribute(0)]];
    float2 attTexcoord0 [[attribute(1)]];
    float3 attNormal [[attribute(2)]];
    float4 attBoneIds [[attribute(3)]];
    float4 attWeights [[attribute(4)]];
};

static inline __attribute__((always_inline))
float2 UV_TRS(thread const float2& inputUV, thread const float2& uvCenter, thread const float2& uvPan, thread const float2& uvScale, thread const float& uvRotate)
{
    float2 minusCenterUV = inputUV - uvCenter;
    float cosVal = cos(uvRotate);
    float sinVal = sin(uvRotate);
    float2x2 rotateMat = float2x2(float2(cosVal, -sinVal), float2(sinVal, cosVal));
    float2 outputUV = minusCenterUV * uvScale;
    outputUV = (rotateMat * outputUV) + uvCenter;
    outputUV += uvPan;
    return outputUV;
}

vertex main0_out main0(main0_in in [[stage_in]], constant buffer_t& buffer)
{
    main0_out out = {};
    float4x4 _79 = buffer.u_Palatte[int(in.attBoneIds.x)] * in.attWeights.x;
    float4x4 _88 = buffer.u_Palatte[int(in.attBoneIds.y)] * in.attWeights.y;
    float4x4 _101 = float4x4(_79[0] + _88[0], _79[1] + _88[1], _79[2] + _88[2], _79[3] + _88[3]);
    float4x4 _110 = buffer.u_Palatte[int(in.attBoneIds.z)] * in.attWeights.z;
    float4x4 _123 = float4x4(_101[0] + _110[0], _101[1] + _110[1], _101[2] + _110[2], _101[3] + _110[3]);
    float4x4 _132 = buffer.u_Palatte[int(in.attBoneIds.w)] * in.attWeights.w;
    float4x4 boneTransform = float4x4(_123[0] + _132[0], _123[1] + _132[1], _123[2] + _132[2], _123[3] + _132[3]);
    float4 bm_postiton = boneTransform * float4(in.attPosition, 1.0);
    float3 bn_normal = (boneTransform * float4(in.attNormal, 0.0)).xyz;
    out.v_posWS = (buffer.u_Model * bm_postiton).xyz;
    out.v_nDirWS = (buffer.u_TransposeInvModel * float4(bn_normal, 0.0)).xyz;
    out.gl_Position = buffer.u_MVP * bm_postiton;
    out.v_uv0 = float2(in.attTexcoord0.x, 1.0 - in.attTexcoord0.y);
    float2 param = out.v_uv0;
    float2 param_1 = float2(0.5);
    float2 param_2 = buffer._Offset;
    float2 param_3 = buffer._Tiling;
    float param_4 = buffer._Rotation / 57.295780181884765625;
    out.v_uv0 = UV_TRS(param, param_1, param_2, param_3, param_4);
    out.gl_Position.z = (out.gl_Position.z + out.gl_Position.w) * 0.5;       // Adjust clip-space for Metal
    return out;
}

