#pragma clang diagnostic ignored "-Wmissing-prototypes"

#include <metal_stdlib>
#include <simd/simd.h>

using namespace metal;

struct buffer_t
{
    float4x4 u_Model;
    float4x4 u_TransposeInvModel;
    float4x4 u_MVP;
    float2 _Offset;
    float2 _Tiling;
    float _Rotation;
};

struct main0_out
{
    float2 v_uv0;
    float3 v_nDirWS;
    float3 v_posWS;
    float4 gl_Position [[position]];
};

struct main0_in
{
    float3 attPosition [[attribute(0)]];
    float2 attTexcoord0 [[attribute(1)]];
    float3 attNormal [[attribute(2)]];
};

static inline __attribute__((always_inline))
float2 UV_TRS(thread const float2& inputUV, thread const float2& uvCenter, thread const float2& uvPan, thread const float2& uvScale, thread const float& uvRotate)
{
    float2 minusCenterUV = inputUV - uvCenter;
    float cosVal = cos(uvRotate);
    float sinVal = sin(uvRotate);
    float2x2 rotateMat = float2x2(float2(cosVal, -sinVal), float2(sinVal, cosVal));
    float2 outputUV = minusCenterUV * uvScale;
    outputUV = (rotateMat * outputUV) + uvCenter;
    outputUV += uvPan;
    return outputUV;
}

vertex main0_out main0(main0_in in [[stage_in]], constant buffer_t& buffer)
{
    main0_out out = {};
    out.v_posWS = (buffer.u_Model * float4(in.attPosition, 1.0)).xyz;
    out.v_nDirWS = (buffer.u_TransposeInvModel * float4(in.attNormal, 0.0)).xyz;
    out.gl_Position = buffer.u_MVP * float4(in.attPosition, 1.0);
    out.v_uv0 = float2(in.attTexcoord0.x, 1.0 - in.attTexcoord0.y);
    float2 param = out.v_uv0;
    float2 param_1 = float2(0.5);
    float2 param_2 = buffer._Offset;
    float2 param_3 = buffer._Tiling;
    float param_4 = buffer._Rotation / 57.295780181884765625;
    out.v_uv0 = UV_TRS(param, param_1, param_2, param_3, param_4);
    out.gl_Position.z = (out.gl_Position.z + out.gl_Position.w) * 0.5;       // Adjust clip-space for Metal
    return out;
}

